#!/usr/bin/perl -w

#Copyright (C) 2006-2008 Keio University
#(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)
#
#This file is part of Murasaki.
#
#Murasaki is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#Murasaki is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with Murasaki.  If not, see <http://www.gnu.org/licenses/>.

use File::Basename;
use Bio::SeqIO;
use Getopt::Long;
use Pod::Usage;

use strict;
my ($help,$man,$opt_prefix);

our ($seqhome,$root,$flexible);
$root=(fileparse($0))[1];
$root=$ENV{MURASAKI} if $ENV{MURASAKI};
require "${root}/config.pl";
require "$root/common.pl";

GetOptions('help|?' => \$help, man => \$man, flexible => \$flexible);
pod2usage(1) if $help or $#ARGV<0;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;

our $out_prefix;
my $alignment_src=shift(@ARGV);
our $format="Murasaki";
our %alignment;
%alignment=%{loadMurasakiAlignment($alignment_src)};
our @seqs=@{$alignment{seqs}};
our @anchors=@{$alignment{anchors}};
print "$format file describes ".($#seqs+1)." sequences and ".($#anchors+1)." anchorss\n";

##read sequences features
foreach my $seq (@seqs){
  my $gb_file=${$seq}{seqFile};
  my @cds_features;
  my ($name,$path,$suffix) = fileparse($gb_file, qr{\.[^.]*});
  if(-e "$gb_file.cds"){
    @cds_features=readCds("$gb_file.cds");
    ${$seq}{seqFeatures}=\@cds_features;
  }else{
    print "$gb_file.cds not found. trying to load from $gb_file\n";
    @cds_features = grep { $_->primary_tag eq 'CDS' }
    Bio::SeqIO->new(-file => $gb_file)->next_seq->get_SeqFeatures;
    print "$gb_file loaded. Contains ".($#cds_features+1)." CDSs\n";
    ${$seq}{seqFeatures}=\@cds_features;
    writeCds("$gb_file.cds",@cds_features);
  }
}

our (@multiname_genes,%allGenes,%quickOrder,%rates);

print "Finding gene Locations...\n";

##find which anchors each feature is on
foreach my $seq (@seqs){
  $$seq{geneLocations}={};
  $$seq{geneanchors}={};
  foreach my $feature (@{$$seq{seqFeatures}}){
    next unless $feature->has_tag('gene');
    my @gene_names=$feature->get_tag_values('gene');

    push(@multiname_genes,@gene_names) if $#gene_names>0;
    foreach my $gene (@gene_names) { #in my data so far there is nothing multiple gene names, but just in case, duplicate them...
      $quickOrder{$gene}=$feature->location->start unless exists $quickOrder{$gene};
      push(@{${$$seq{geneLocations}}{$gene}},$feature->location);
      my @usedanchors=findanchors($$seq{seqId},$feature->location);
      push(@{${$$seq{geneanchors}}{$gene}},@usedanchors);
      ${$allGenes{$gene}}[$$seq{seqId}]=[@usedanchors]; #fill matrix
    }
  }
}

if($#multiname_genes>=0){
  writeOut("$out_prefix.multinames",join("\n",@multiname_genes))
}

print "Sorting...\n";
our @genes=sort(geneSort keys(%allGenes));
my $allGeneCount=$#genes+1;
#total gene list
print "Writing out complete list of $allGeneCount genes...\n";
writeOut("$out_prefix.genelist",prettyGenePrint(@genes));

our (%active,%histo,%phisto,%rhit,%rhisto,%missed,%matched);
#compute histograms for different multiplicity classes
for(1..mclassToI('all')){
  my @m=mclass($_);
  my $m=mclassToStr(@m);
#  print "Selecting mclass $_(".mclassToStr(@m).")...\n";
  print "mclass $_(".mclassToStr(@m)."): ";
  my @active=grep({mclassMember($_,@m)} @genes);
  my @matched=grep({mclassMatched($_,@m)} @genes);
  my @missed=grep({!mclassMatched($_,@m)} @active);
  $active{$m}=$#active+1;
  $matched{$m}=$#matched+1;
  $missed{$m}=$#matched+1;
  print sprintf("%d (%2d%%) gene-name cover, %d (%d%%) active\n",
		$active{$m},$active{$m}/$allGeneCount*100,
		$matched{$m},$matched{$m}/$active{$m}*100);
  next unless $#active>=0;
  writeOut("$out_prefix.$m",prettyGenePrint(@active));
  writeOut("$out_prefix.$m.miss",missAnalysis(\@m,\@missed));
  if(magnitude(@m)>1) { #bits to compare to!
    my @aligned=grep {aligned($_,@m)} @matched;
    my @missAligned=grep {!aligned($_,@m)} @matched;
    writeOut("$out_prefix.$m.missalign",prettyGenePrint(@missAligned));
    my @partiallyAligned=grep {partiallyAligned($_,@m)} @matched;
    my @missPartiallyAligned=grep {!partiallyAligned($_,@m)} @matched;
    writeOut("$out_prefix.$m.missalign-partial",prettyGenePrint(@missPartiallyAligned));
    $histo{$m}=$#aligned+1;
    $phisto{$m}=$#partiallyAligned+1;
    print "\t==>$histo{$m}(".($histo{$m}/$matched{$m}*100)."%) genes remain aligned to the same anchor.\n".
      "\t==-->$phisto{$m}(".($phisto{$m}/$matched{$m}*100)."%) contain at least one matching anchor\n";
  }else{
    $histo{$m}=$matched{$m}; #by definition
  }
}
writeOut("$out_prefix.stats",
	 join("\n",
	      join("\t",getName($alignment_src),"Total anchors:",$#anchors+1),
	      join("\t",qw{Class Mag Genes Genes% inanchor inanchor% Part Part% Match Match%}),
	      (map {
		my @m=mclass($_);
		my $m=mclassToStr(@m);
		join("\t",$m,magnitude(@m),
		     #cover
		     $active{$m},
		     sprintf("%.2f%%",$active{$m}/$allGeneCount*100),
		     #matched
		     $matched{$m},
		     sprintf("%.2f%%",$matched{$m}/$active{$m}*100),
		     #partial
		     $phisto{$m},
		     sprintf("%.2f%%",$phisto{$m}/$matched{$m}*100),
		     #aligned
		     $histo{$m},
		     sprintf("%.2f%%",$histo{$m}/$matched{$m}*100))
	      } grep {magnitude(mclass($_))>1}
	      (1..mclassToI('all'))),""
	     ));

exit 0;

sub missAnalysis {
  my @m=@{scalar(shift)};
  my @slice=mclassToSlice(@m);
  my @misses=@{scalar(shift)};
  my $ret=join("\t",qw{Gene Avg StdDev},map {"seq$_"} @slice)."\n";
  foreach my $gene (@misses){
    my @lengths=map {geneLength($gene,$_)} @slice;
    $ret.=join("\t",$gene,trunc(avg(@lengths)),trunc(stddev(@lengths)),@lengths)."\n";
  }
  return $ret;
}

sub trunc {
  map {sprintf("%.2f",$_)} @_;
}

sub geneLength {
  my ($gene,$seqid)=(@_);
  my $seqr=$seqs[$seqid];
  my $loc=${${$$seqr{geneLocations}}{$gene}}[0];
  return 0 unless ref($loc);
  my @coords=toMauveCoords($loc);
  return $coords[1]-$coords[0];
}

sub sum {
  my $sum;
  for(@_){$sum+=$_}
  return $sum;
}

sub avg {
  return sum(@_)/($#_+1);
}

sub sq {
  my $a=pop;
  return $a*$a;
}

sub stddev {
  our $avg=avg(@_);
  return sqrt(avg(map {sq($_-$avg)} @_));
}

sub getBucket {
  my $val=shift;
  return int($val*100);
}

sub magnitude {
  return scalar(grep({$_} @_));
}

sub mclassToSlice {
  my @l;
  for(0..$#_){
    push(@l,$_) if $_[$_];
  }
  return @l;
}

sub aligned {
  my ($gene,@m)=@_;
  my @used_rl=@{$allGenes{$gene}}[mclassToSlice(@m)];
  my @ref=map {$$_{id}} @{$used_rl[0]};
  foreach my $usedr (@used_rl[1..$#used_rl]){
    return 0 if $#{$usedr}!=$#ref;
    foreach my $anchori (0..$#ref){
      return 0 if $ref[$anchori]!=$$usedr[$anchori]{id};
    }
  }
  return 1;
}

sub partiallyAligned {
  my ($gene,@m)=@_;
  my @used_rl=@{$allGenes{$gene}}[mclassToSlice(@m)];
  my @ref=map {$$_{id}} @{$used_rl[0]};
  foreach my $usedr (@used_rl[1..$#used_rl]){
    foreach my $anchori (0..$#ref){
#      print "$gene - $#$usedr\n";
      return 1 if grep({$ref[$anchori]==$$usedr[$_]{id}} 0..$#$usedr);
    }
  }
  return 0;
}

sub prettyGenePrint {
  join("\n",map {
    my ($gene)=$_;
    join("\t",$_,
	 map {my $seqId=$_; 
	      join(",",map {
		  my $usedr=$_;
		  "$$usedr{id}.$$usedr{partial}"
		  } @{${$allGenes{$gene}}[$seqId]})
	      } (0..$#seqs)) } @_ )
}

sub geneSort {
  return $quickOrder{$a}<=>$quickOrder{$b};
}

sub pow {
  my ($b,$p)=(shift,(shift)-1);
  my $r=$b;
  for (1..$p){
    $r*=$b;
  }
  return $r;
}

sub digits {
  my ($a,$b)=@_;
  my @l;
  while($a>0){
    unshift(@l,$a % $b);
    $a=int($a / $b);
  }
  return @l;
}

sub pad {
  my $fill=shift;
  my $target=shift;
  unshift(@_,$fill) while($#_+1<$target);
  return @_;
}

sub mclassToStr {
  return join("",@_);
}

sub mclass {
  my $i=shift;
  $i=pow(2,$#seqs+1)-1 if $i=~m/all/i;
  return pad(0,$#seqs+1,digits($i,2));
}

sub mclassToI {
  my $i=0;
  return pow(2,$#seqs+1)-1 if $_[0]=~m/all/i;
  $i=($i+pop) << 1 while(@_);
  return $i>>1;
}

sub mclassMember {
  my $gene=shift;
  my $usedr=$allGenes{$gene};
  return !grep {my $seqid=$_;my $seqr=$seqs[$seqid];
		$_[$_] and $#{$$seqr{geneLocations}{$gene}}<0} (0..$#_);
}
#wth was this for?
#  return !grep {!((ref(${$usedr}[$_]) and $_[$_]) or !$_[$_])} (0..$#_);

sub mclassMatched {
  my $gene=shift;
  my $usedr=$allGenes{$gene};
  return !grep {$_[$_] and $#{${$usedr}[$_]}<0} (0..$#_);
}

sub toMauveCoords {
  my @coords=($_[0]->start,$_[0]->end);
  return @coords;
}

sub findanchors {
  my ($genome,$start,$stop)=($_[0],toMauveCoords($_[1]));
#  print "Searching ".($#anchors+1)." anchors on genome $genome\n";
#  print join("\n",map {"$_ -> ".ref($anchors[$_])."=".join(" ",@{$anchors[$_]})} 0..$#anchors);
  return grep {grep {$_} $$_{partial}}
    (map {my @anchorl=@$_;
	  {partial=>coversanchor($start,$stop,$anchorl[$genome]), id=>$anchorl[$genome]{anchorId}} } @anchors);
}

#returns 0 or which side of pair 1 (gene coords) overhangs pair 2 (anchor)
sub coversanchor {
  my ($start,$stop,$anchorr)=@_;
  my @a=($start,$stop,$$anchorr{start},$$anchorr{stop});
  return covers(@a);
}

sub covers {
  my ($a,$b,$c,$d)=@_;
  $a<=$b or die "start<=stop assertion failed";
  $c<=$d or die "anchor start<=stop assertion failed";
  return 0 if $a>$d or $c>$b; # total mismatch
  #  if($a<=$d and $c<=$b){ #some degree of hit (already guranteed by above)
  return "none" if $c<=$a and $b<=$d;
  return "left" if $c<=$a and $d<=$b;
  return "right" if $a<=$c and $b<=$d;
  return "both" if $a<=$c and $d<=$b;
}

sub loadMauveAlignment {
  my $alignment=shift;
  open(MAUVE,"<$alignment");
  <MAUVE>=~m/FormatVersion\s+(\d+)/ or die "Not a mauve file: $alignment";
  my $version=$1;
  my @seqs=();
  do {print "This program is written for Mauve Format Version 4.\n
This file is version $version. Weird stuff may happen.\n"; $flexible=1;} if $version!=4;
  <MAUVE>=~m/SequenceCount\s+(\d+)/ or die "Unknown sequence count\n";
  my $seqCount=$1;
  while(<MAUVE>){
    next unless m/Sequence(\d+)File\s+(\S.*)/;
    my ($seqId,$seqFile)=($1,$2);
    $_=<MAUVE>;
    m/Sequence${seqId}Length\s+(\d+)/ or $flexible or die "Input file is weird: $_";
    my $seqLength=$1;
    $seqs[$seqId]={'seqId' => $seqId,'seqFile' => $seqFile,'seqLength'=>$seqLength};
    last if $seqId==$seqCount-1;
  }

  @anchors=();
  $_=<MAUVE>;
  m/IntervalCount\s(\d+)/ or $flexible or die "Interval Count line weird: $_";
  my $anchorCount=$1;
  while(<MAUVE>){
    m/Interval\s(\d+)/ or next;
    my $anchorId=$1;
    $_=<MAUVE>;
    chomp;
    my ($length,@start)=split(/\s+/);
    @start=map(abs,@start);
    my @stop=map {$_+$length} @start;
    my @segs;
    while(<MAUVE>){
      chomp;
      last if $_ eq '';
      next if $_ eq 'GappedAlignment' or m/^[A-Z-]+$/; #skip gapped lines
      ($length,@segs)=split(/\s+/);
      @stop=map {$_+$length} @segs;
    }
    next if (grep {$_==0} @start)>0;
    my @anchor=map {
      ($start[$_],$stop[$_])=map(abs,($stop[$_],$start[$_])) if $start[$_]>$stop[$_]; #is rev strand?
      { start => $start[$_], stop => $stop[$_], anchorId => $anchorId }} 0..$#stop;
    
    push(@anchors,\@anchor);
  }
  return {'seqs' => \@seqs, 'anchors' => \@anchors};
}

sub getName {
  my @ret=map {
    my ($name,$path,$suffix) = fileparse($_, qr{\.[^.]*});
    $name
    } @_;
  return @ret if $#_;
  return $ret[0];
}

sub loadMurasakiAlignment {
  my $filename=pop;
  my @seqs;
  if(-e "$filename.seqs"){
    my $seqId=0;
    open(SEQS,"$filename.seqs") or die "Could not open seqs file";
    while(<SEQS>){
      chomp;
      push(@seqs,{seqId => $seqId++, seqFile => $_});
    }
    close(SEQS);
  }else{
    my @seqnames=split(/-/,getName($filename));
    my $seqId=0;
    foreach(@seqnames){
      my $name="$seqhome/$_.gbk";
      -e $name or die "Sequence $name not found.\nEither use correct naming style, or create $filename.seqs file specifying sequence locations.";
      push(@seqs,{seqId => $seqId++, seqFile => $name});
    }
  }

  my @anchors;
  my $anchorId=0;
  open(BLOCKS,$filename) or die "Blocks file not found...??";
  while(<BLOCKS>){
    chomp;
    $anchorId++;
    my @anchor=();
    while(m/(\d+) (\d+) ([+-])/g){
      my ($start,$stop)=($1,$2);
      ($start,$stop)=map(abs,($stop,$start)) if $start>$stop; #is rev strand?
      push(@anchor,{start => $start, stop => $stop, anchorId => $anchorId});
    }
    push(@anchors,\@anchor);
  }
  return {'seqs' => \@seqs, 'anchors' => \@anchors};
}


__END__

=head1 NAME

cbtest.pl - Conserved Block Test

=head1 SYNOPSIS

cbtest.pl [-output=<ouput prefix>] <murasaki anchors file>

=head1 PARAMETERS

=over 8

=item B<murasaki anchors file>
The main anchors file outputted by Murasaki

For example:
	cbtest.pl myAlignment.anchors
Results in files like: myAlignment.block.stats etc...

=back

=head1 DESCRIPTION

B<This program> uses annotation data in genebank files to assess
how well Murasaki correctly aligned regions.

=cut
